# -*- coding: utf-8 -*-

'''USAQG

SDFI jasodfij aosdif a
'''

import pika
import os
import json
import time
import itertools


class ZZNBus(object):

    def __init__(self, url, rpcExchange='RPC', eventExchange='EVT'):
        self.rpcExchange = rpcExchange
        self.eventExchange = eventExchange
        self.conn = pika.BlockingConnection(pika.URLParameters(url))
        self._rpc = None
        self._sendChannel = None

    @property
    def sendChannel(self):
        if self._sendChannel is None:
            self._sendChannel = self.conn.channel()
        return self._sendChannel

    def send(self, routing_key, event):
        self.sendChannel.basic_publish(
            exchange=self.eventExchange,
            routing_key=routing_key,
            body=json.dumps(event),
            mandatory=False,
        )

    def monitor(self, bindings, callback):
        chan = self.conn.channel()
        queue = chan.queue_declare(exclusive=True).method.queue
        for exchange, routing_key in bindings:
            print 'exchange=%s routing_key=%s' % (exchange, routing_key)
            chan.queue_bind(
                queue=queue,
                exchange=exchange,
                routing_key=routing_key,
                # nowait=True,
                # arguments=None
            )
        chan.basic_consume(
            consumer_callback=callback,
            queue=queue,
            no_ack=True,
            exclusive=True,
            # consumer_tag=None,
            # arguments=None
        )
        print 'start receiving messages...'
        chan.start_consuming()

    def close(self):
        self.conn.close()
        if self._rpc is not None:
            self._rpc.close()

    def startConsumer(self, bindings, queueName, callback):
        chan = self.conn.channel()
        chan.queue_declare(
            queue=queueName,
            passive=False,
            durable=False,
            exclusive=False,
            auto_delete=True,
            nowait=False,
            arguments=None
        )
        for rk in bindings:
            chan.queue_bind(
                queue=queueName,
                exchange=self.eventExchange,
                routing_key=rk,
                nowait=False,
                arguments=None
            )

        def consumer(ch, method, props, body):
            body = json.loads(body)
            callback(method.routing_key, body)
        chan.basic_consume(consumer, queue=queueName, no_ack=False)
        chan.start_consuming()

    def startRPC(self, methods, queueName, qos=1):
        chan = self.conn.channel()
        if qos:
            chan.basic_qos(prefetch_count=qos)
        chan.queue_declare(
            queue=queueName,
            passive=False,
            durable=False,
            exclusive=False,
            auto_delete=True,
            nowait=False,
            arguments=None
        )
        for name in methods:
            chan.queue_bind(
                queue=queueName,
                exchange=self.rpcExchange,
                routing_key=name,
                nowait=False,
                arguments=None
            )

        def consumer(ch, method, props, body):

            req = reqid = None
            name = method.routing_key
            cid = props.correlation_id
            # print 'REQ', method, props, body

            def reply(result=None, error=None):
                # print 'REPLY', (result, error)
                rep = dict(id=reqid)
                if result is not None:
                    rep['result'] = result
                if error is not None:
                    rep['error'] = error
                chan.basic_publish(
                    exchange="",
                    routing_key=props.reply_to,
                    properties=pika.BasicProperties(correlation_id=cid),
                    body=json.dumps(rep),
                    mandatory=False,
                    immediate=False
                )
            try:
                req = json.loads(body)
                reqid = req.get('id')
                if name not in methods:
                    reply(error='Method %s not implemented' % name)
                res = methods[name](*req['params'])
                reply(result=res)
            except Exception, e:
                reply(error='%s: %s' % (e.__class__.__name__, str(e)))
            chan.basic_ack(delivery_tag=method.delivery_tag)
        chan.basic_consume(consumer, queue=queueName, no_ack=False)
        chan.start_consuming()

    @property
    def rpc(self):
        if self._rpc is None:
            self._rpc = RPCClient(self)
        return self._rpc


class RPCClient(object):

    def __init__(self, bus):
        self._bus = bus
        self._chan = bus.conn.channel()
        queue = self._chan.queue_declare(exclusive=True)
        self._queue = queue.method.queue
        self._ctag = self._chan.basic_consume(
            self._consumer, queue=self._queue, no_ack=True
        )
        self._results = {}

    def close(self):
        self._chan.close()

    def _consumer(self, ch, method, props, body):
        self._results[props.correlation_id] = json.loads(body)

    _idCounter = itertools.count()

    def nextCorrelationId(self):
        return 'req-%d-%s-%d' % (
            os.getpid(), id(self), self._idCounter.next()
        )

    def call(self, requests, timeout=300):
        assert timeout is not None and timeout > 0
        start = time.time()
        results = {}
        reqmap = {}
        cid = self.nextCorrelationId()
        props = pika.BasicProperties(
            reply_to=self._queue,
            correlation_id=cid,
            expiration = str((timeout + 2) * 1000),
        )
        for request in requests:
            request['id'] = cid
            reqmap[cid] = request
            self._chan.basic_publish(
                exchange=self._bus.rpcExchange,
                routing_key=request['method'],
                properties=props,
                body=json.dumps(request),
                mandatory=True,
            )
        # Collect responses
        while timeout is None or timeout > (time.time() - start):
            for cid, res in self._results.items():
                # We can receive "fantom" replay from previous call that
                # didn't finished within given timeout
                if cid not in reqmap:
                    continue
                results[cid] = res
            self._results.clear()
            if len(results) == len(requests):
                break
            self._bus.conn.process_data_events()
        # Create Timeout response for requests that didnt finished
        for cid, request in reqmap.items():
            if cid not in results:
                results[cid] = dict(
                    id=cid,
                    error="Timeouted after %s seconds" % timeout,
                    result=None,
                )
        # Result
        return [(request, results[cid]) for cid, request in reqmap.items()]

    def __call__(self, request):
        res = self.call([request])
        return res[0][1]

    def __getattr__(self, name):
        return RpcNamespace(self, name)


class RpcNamespace(object):

    def __init__(self, rpc, name):
        self.__rpc = rpc
        self.__name = name

    def __getattr__(self, method):
        return RpcNamespace(self.__rpc, '%s.%s' % (self.__name, method))

    def __call__(self, *args, **kwds):
        if args:
            assert not kwds, repr((args, kwds))
            params = list(args)
        elif kwds:
            params = [kwds]
        else:
            params = []
        request = dict(
            method=self.__name,
            params=params
        )
        res = self.__rpc(request)
        if res.get('error'):
            raise Exception(res['error'])
        else:
            return res.get('result')


DEFAULT_URL = "amqp://agent:{MQ_PASSWD}@mq.is.uh.in/dat"


def main():
    from argparse import ArgumentParser
    # from agent5.loggers import script_logging

    parser = ArgumentParser()
    parser.add_argument(
        '-v', '--verbose', default=0, action='count')
    parser.add_argument(
        '-U', '--url', metavar='URL', default=DEFAULT_URL,
        help='RabbitMQ URL (default %(default)s)')
    parser.add_argument(
        'action', metavar='ACTION',
        help='Action')
    parser.add_argument(
        'args', nargs='*', metavar='ARG',
        help='Action arguments')

    args = parser.parse_args()
    args.url = args.url.format(**os.environ)
    # script_logging(verbosity=args.verbose)
    # logging.getLogger('pika').setLevel(logging.WARNING)

    if args.action == 'shell':
        if args.args:
            parser.error('unwanted args for shell action')
        return action_shell(args.url)
    elif args.action == 'monitor':
        if not args.args:
            bindings = {'EVT': ['#']}
        else:
            bindings = [arg.split(':', 1) for arg in args.args]
        return action_monitor(args.url, bindings)
    else:
        parser.error('unknown action: %s' % args.action)
    return 0


def action_monitor(url, bindings):

    def dump_message(ch, method, props, body):
        print '------------------------------------------------------------'
        print '%s -> %s %s' % (
            method.exchange,
            method.routing_key,
            json.dumps(json.loads(body), indent=2, sort_keys=True)
        )
    bus = ZZNBus(url)
    bus.monitor(bindings, dump_message)


def action_shell(url, histfile='~/.rpc-shell-history'):
    import code
    import sys
    import pprint
    import rlcompleter
    import readline
    import atexit
    sys.displayhook = pprint.pprint
    bus = ZZNBus(url)
    namespace = dict(bus=bus)
    readline.set_completer(rlcompleter.Completer(namespace).complete)
    readline.parse_and_bind("tab: complete")
    if histfile:
        histfile = os.path.expanduser(histfile)
        if os.path.isfile(histfile):
            readline.read_history_file(histfile)

        def savehist():
            readline.set_history_length(100)
            readline.write_history_file(histfile)

        atexit.register(savehist)
    con = code.InteractiveConsole(namespace)
    con.resetbuffer()
    con.interact(banner='''
bus = ZZNBus instance (connected)
''')
